-- Requirements
-- * Work with Hotpatch
-- x Toggelable in-game timer for players
-- * Spawn biters at map edge when timer runs out
--   x Tile walk out from spawn until water is found
--   x Walk along the shore once water is found
--   x spawn biters where there are room
--   x work around biter path issues
--   - destroy player structers along the shore (to avoid cheese)
--   x Make the biters much tougher (to compensate for tech sync)
--     - May not actually be neccessary
--     - Nerfing the bonuses is much simpler
--
-- * win condition
-- * in-game config that can be accessed by admins

local mod_gui = require("mod-gui")

-- This is based heavily on the silo-script
local script_data = {
    
    -- Time before the wrath starts
    time_left = 10,

    -- Number of rays to shoot out at wrath start
    ray_count = 7,
    
    -- Number of tiles to walk rays per wrath tick
    ray_walk_distance = 10,
    
    -- Distance from 0, 0 rays start at
    ray_start_distance = 200,

    -- spawn group fields:
    -- name = Prototype name of unit to spawn
    -- weight = Relative occurence this group will be spawned with
    -- spawn_attempts = The number of spawns to attempt in each group
    -- minimum_group_size = Minimum number of successful spawns in a group
    -- distance_from_shore = The distance towands land spawns will be attempted
    -- separation = The distance along the shore between each spawn
    spawn_groups = {
        {
            name = "behemoth-biter",
            weight = 20,
            spawn_attempts = 20,
            minimum_group_size = 4,
            distance_from_shore = 2,
            separation = 3,
        },
        {
            name = "behemoth-spitter",
            weight = 10,
            spawn_attempts = 20,
            minimum_group_size = 4,
            distance_from_shore = 2,
            separation = 3,
        }
    },
    
    -- Stop sawning when this number of units are on the map
    maximum_concurrent_units = 2000,
    
    -- The area biters spawned will target
    biter_target = {x = 0, y = 0},
    
    -- Draws a lot of debugging graphics
    debug_draw = true,
    
    -- The following are internal variables, do not modify
    version = 1,

    won_time = nil,
    wrath = false,
    dead_biters = {},
    rays = {},
    current_shore_path = 1,
    shore_paths = {},
    unit_groups = {},
}

-- Converts number of seconds to h:mm:ss
local function seconds_to_string(seconds)
    local prefix = ''
    if seconds < 0 then
        seconds = -seconds
        prefix = '-'
    end
    
    local hours = math.floor(seconds / 3600)
    seconds = seconds % 3600
    local minutes = math.floor(seconds / 60)
    seconds = seconds % 60
    
    return string.format('%s%d:%02d:%02d', prefix, hours, minutes, seconds)
end

local function gui_update(player)
    local gui = mod_gui.get_frame_flow(player)
    local frame = gui.poseidon_gui_frame
    if not frame then
        return
    end
    
    frame.clear()
    frame.add{
        type = "label",
        caption = seconds_to_string(script_data.time_left)
    }
end


local function toggle_frame(player)
    local gui = mod_gui.get_frame_flow(player)
    local frame = gui.poseidon_gui_frame
    
    if frame then
        frame.destroy()
        return
    end
    
    frame = gui.add{
        type = "frame",
        name = "poseidon_gui_frame",
        direction = "vertical",
        caption = "Time Left",
        style = mod_gui.frame_style,
    }
    
    frame.style.horizontally_stretchable = false
    frame.style.vertically_stretchable = false
    gui_update(player)
end

local function get_sprite_button(player)
    local button_flow = mod_gui.get_button_flow(player)
    local button = button_flow.poseidon_timer_button
    if not button then
        button = button_flow.add{
            type = "sprite-button",
            name = "poseidon_timer_button",
            sprite = "utility/time_editor_icon",
            style = mod_gui.button_style,
            tooltip = "This button tooltip",
        }
    end
    
    button.visible = true
end

local function on_gui_click(event)
    local gui = event.element
    local player = game.players[event.player_index]
    if not (player and player.valid and gui and gui.valid) then
        return
    end
    
    if gui.name == "poseidon_timer_button" then
        toggle_frame(player)
    end
end

local function on_player_created(event)
    local player = game.players[event.player_index]
    if not (player and player.valid) then
        return
    end
    
    get_sprite_button(player)
    toggle_frame(player)
end

local function gui_update_all()
    for _, player in pairs(game.players) do
        if player and player.valid then
            gui_update(player)
        end
    end
end

local function on_rocket_launched(event)
    if script_data.won_time then
        return
    end
    
    -- First rocket being sent up, set game was won
    script_data.won_time = script_data.time_left
    game.print(
        "Satelite scans have located Poseidon",
        {r=1, g=1, b=0.3}
    )
end


local function on_gui_tick(event)
    if script_data.debug_draw then
        rendering.draw_circle{
            color = {r=1,g=0.5,b=0.2},
            radius = script_data.ray_start_distance,
            width = 5,
            filled = false,
            target = {0, 0},
            surface = 1,
            time_to_live = 61,
        }
    end
    
    if script_data.won_time then
        -- Game ended
        if script_data.won_time < -10 then
            return

        -- Poseidon didn't say a word
        elseif script_data.won_time >= 11 then
            script_data.won_time = -9

        -- Poseidon has said something
        elseif script_data.won_time >= 0 then
            script_data.won_time = -1
        end
        
        if script_data.won_time == -3 then        
            game.print(
                "Poseidon: You win this time",
                {r=0, g=0.8, b=1}
            )
        
        elseif script_data.won_time == -6 then
            game.print(
                "Poseidon: But you won't get away so easily next time",
                {r=0, g=0.8, b=1}
            )
        elseif script_data.won_time == -9 then
            game.print(
                "Congratulation, you won this round.  Poseidon will not dare to come near you"
            )
        end
        
        script_data.won_time = script_data.won_time - 1
        return
    end

    -- Display dialogue for when the timer is about to run out
    if script_data.time_left == 11 then
        game.print(
            "Poseidon: What have you done?",
            {r=0, g=0.8, b=1}
        )
    
    elseif script_data.time_left == 8 then
        game.print(
            "Poseidon: You.  You destroyed my sacred island and killed my people.",
            {r=0, g=0.8, b=1}
        )
    
    elseif script_data.time_left == 1 then
        game.print(
            "Poseidon: FEEL MY WRATH!",
            {r=0, g=0.8, b=1}
        )
    end
    
    -- Count down until 0
    if script_data.time_left > 0 then
        script_data.time_left = script_data.time_left - 1
        
        -- Start the wrath when the time runs out
        if script_data.time_left == 0 then
            script_data.wrath = true
        end
        
        gui_update_all()
    end
end

local water_tiles = {
    ["water"] = true,
    ["deepwater"] = true,
    ["water-green"] = true,
    ["deepwater-green"] = true,
    ["out-of-map"] = true,
}

local function is_water(tile)
    if not tile.valid then
        return true
    end
    return water_tiles[tile.name] == true
end

local function to_direction(dx, dy)
    local magic = math.cos(math.pi / 8)
    if dx > 0 then
        if dx > magic then return defines.direction.east
        elseif dy > 0 then
            if dy > magic then return defines.direction.south
            else return defines.direction.southeast end
        else
            if dy < -magic then return defines.direction.north
            else return defines.direction.northeast end
        end
    else
        if dx < -magic then return defines.direction.west
        elseif dy > 0 then
            if dy > magic then return defines.direction.south
            else return defines.direction.southwest end
        else
            if dy < -magic then return defines.direction.north
            else return defines.direction.northwest end
        end
    end
end

local function add_direction(position, direction)
    local new_position = {x = position.x, y = position.y}
    if -- Note, this wraps around, northwest is 7, northeast is 1
        direction >= defines.direction.northwest
        or direction <= defines.direction.northeast
    then
        new_position.y = new_position.y - 1
    end
    
    if
        direction >= defines.direction.southeast
        and direction <= defines.direction.southwest
    then
        new_position.y = new_position.y + 1
    end
    
    if
        direction >= defines.direction.northeast
        and direction <= defines.direction.southeast
    then
        new_position.x = new_position.x + 1
    end
    if
        direction >= defines.direction.southwest
        and direction <= defines.direction.northwest
    then
        new_position.x = new_position.x - 1
    end
    
    return new_position
end

function walk_ray(ray)
    local tile = game.surfaces[1].get_tile(ray.x, ray.y)
    local found_water = is_water(tile)
    
    if
        ray.previous_was_water == false and found_water == true
        or ray.previous_was_water == true and found_water == false
    then
        local prev =
            game.surfaces[1].get_tile(ray.x - ray.dx, ray.y - ray.dy)
        table.insert(script_data.shore_paths, {
            position = prev.position,
            direction = to_direction(ray.dx, ray.dy),
        })
        return true

    -- Flip arround in case we start in water
    elseif ray.previous_was_water == nil and found_water == true then
        ray.dx = -ray.dx
        ray.dy = -ray.dy
        ray.previous_was_water = found_water
        
    else
        if script_data.debug_draw then
            rendering.draw_rectangle{
                color = {r=0.7,g=0.7,b=0.7},
                width = 5,
                filled = false;
                left_top = tile.position,
                right_bottom = {
                    x = tile.position.x + 1,
                    y = tile.position.y + 1,
                },
                surface = 1,
                time_to_live = 120,
            }
        end
        
        ray.x = ray.x + ray.dx
        ray.y = ray.y + ray.dy
        ray.previous_was_water = found_water
    end
    
    return false
end

-- Progress one tile along the shore clockwise.  Returns false
-- if the function was unable to locate the shore.
local function follow_shore(path)
    local new_direction = path.direction
    local new_position = add_direction(path.position, new_direction)
    local new_tile = game.surfaces[1].get_tile(new_position)
    
    if is_water(new_tile) then -- Rotate clockwise until land
        local found_land = false
        for i=1,8 do
            new_direction = (new_direction + 1) % 8
            new_position = add_direction(path.position, new_direction)
            new_tile = game.surfaces[1].get_tile(new_position)
            
            if not is_water(new_tile) then
                found_land = true
                break
            elseif script_data.debug_draw then
                rendering.draw_rectangle{
                    color = {r=0,g=0.1*i,b=1},
                    width = 2,
                    filled = false;
                    left_top = new_position,
                    right_bottom = {
                        x = new_position.x + 1,
                        y = new_position.y + 1,
                    },
                    surface = 1,
                    time_to_live = 120,
                }
            end
        end
        
        if not found_land then
            return false
        end
        
    else -- Rotate counter clockwise until water
        local found_water = false
        local land_tile = new_tile
        for i=1,8 do
            new_direction = (new_direction + 7) % 8
            new_position = add_direction(path.position, new_direction)
            new_tile = game.surfaces[1].get_tile(new_position)
            
            if is_water(new_tile) then
                found_water = true
                break
            elseif script_data.debug_draw then
                rendering.draw_rectangle{
                    color = {r=0,g=1,b=0.1*i},
                    width = 2,
                    filled = false;
                    left_top = new_position,
                    right_bottom = {
                        x = new_position.x + 1,
                        y = new_position.y + 1,
                    },
                    surface = 1,
                    time_to_live = 120,
                }
            end
            
            land_tile = new_tile
        end
        
        if not found_water then
            return false
        end
        
        new_direction = (new_direction + 1) % 8
        new_tile = land_tile
    end
    
    path.position = new_tile.position
    path.direction = new_direction
    
    if script_data.debug_draw then
        rendering.draw_rectangle{
            color = {r=1,g=1,b=1},
            width = 2,
            filled = false;
            left_top = path.position,
            right_bottom = {
                x = path.position.x + 1,
                y = path.position.y + 1,
            },
            surface = 1,
            time_to_live = 120,
        }
    end
    
    return true
end

local function calculate_spawn(start_position, end_position, separation)
    local midpoint = {
        x = (start_position.x + end_position.x) / 2,
        y = (start_position.y + end_position.y) / 2,
    }
    
    local vector = {
        x = end_position.x - start_position.x,
        y = end_position.y - start_position.y,
    }
    local length = math.sqrt(vector.x * vector.x + vector.y * vector.y)
    
    -- Start and end position may be equal to each other
    if length < 0.1 then
        return midpoint
    end
    
    vector.x = vector.x / length
    vector.y = vector.y / length
    
    local rotated = {
        x = -vector.y,
        y = vector.x,
    }
    
    return {
        x = midpoint.x + rotated.x * separation,
        y = midpoint.y + rotated.y * separation,
    }
end

local function pick_spawn_group()
    local total_weight = 0
    for i,group in ipairs(script_data.spawn_groups) do
        total_weight = total_weight + group.weight
    end
    
    local target_weight = math.random() * total_weight
    for i,group in ipairs(script_data.spawn_groups) do
        if target_weight < group.weight then
            return group
        end
        target_weight = target_weight - group.weight
    end
    
    -- Shouldn't be possible to reach here
    return script_data.spawn_groups[1]
end

-- XXX GROSE DEBUG
local function find_all_enemy_unit_groups()
    local units = game.surfaces[1].find_entities_filtered{
        type = "unit",
        force = "enemy",
    }
    
    local groups = {}
    local count = 0
    for i,unit in ipairs(units) do
        local unit_group = unit.unit_group
        if not unit_group then
            count = count + 1
        elseif not groups[unit_group.group_number] then
            groups[unit_group.group_number] = unit_group
        end
    end
    
    -- XXX DEBUG
    --if count > 0 then
    --    game.print("Groupless biters: "..count)
    --end
    return groups
end

--- XXX DEBUG
local function on_every_tick(event)
    if script_data.debug_draw then
        for id,group in pairs(script_data.unit_groups) do
            if group.valid then
                local color
                if group.state == defines.group_state.gathering then
                    color = {r=0,g=0.5,b=1}
                elseif group.state == defines.group_state.moving then
                    color = {r=0,g=1,b=0}
                elseif group.state == defines.group_state.attacking_distraction then
                    color = {r=1,g=1,b=0}
                elseif group.state == defines.group_state.attacking_target then
                    color = {r=1,g=0.5,b=0}
                elseif group.state == defines.group_state.finished then
                    color = {r=1,g=1,b=1}
                elseif group.state == defines.group_state.pathfinding then
                    color = {r=0,g=0,b=0}
                elseif group.state == defines.group_state.wander_in_group then
                    color = {r=0,g=1,b=1}
                else
                    color = {r=1,g=0,b=1}
                end
                
                rendering.draw_circle{
                    color = color,
                    radius = 4.5,
                    width = 5,
                    filled = false,
                    target = group.position,
                    surface = 1,
                    time_to_live = 2,
                }
                
                local text = string.format(
                    "id %d members %d state %d, pos %s",
                    id, #group.members, group.state, serpent.block(group.position)
                )
                rendering.draw_text{
                    text = text,
                    surface = 1,
                    target = group.position,
                    color = {r=1,g=1,b=1},
                    time_to_live = 2,
                }
            end
        end
    end
end

local function on_wrath_tick(event)
    -- This counting costs about 0.1ms on a small map.
    if 
        not script_data.wrath
        or script_data.won_time
        or game.surfaces[1].count_entities_filtered{
            type = "unit",
            force = "enemy",
        } > script_data.maximum_concurrent_units
    then
        if script_data.debug_draw then
            rendering.draw_circle{
                color = {r=1,g=0,b=0},
                radius = 20,
                width = 20,
                filled = false,
                target = {0, 0},
                surface = 1,
                time_to_live = 25,
            }
        end

        return
    end
    
    -- Clean up unit_groups
    for id,group in pairs(script_data.unit_groups) do
        if not group.valid then
            script_data.unit_groups[id] = nil
        end
    end
    
    if script_data.ray_count then
        local ray_count = script_data.ray_count
        script_data.ray_count = nil
        
        for i=1,ray_count do
            local angle = math.pi * 2 / ray_count * (i - 1)
            local ray = {
                dx = math.sin(angle),
                dy = math.cos(angle),
                x = math.sin(angle) * script_data.ray_start_distance,
                y = math.cos(angle) * script_data.ray_start_distance,
            }

            table.insert(script_data.rays, ray)
        end
    end
    
    if #script_data.rays > 0 then
        local done = {}
        for i,ray in ipairs(script_data.rays) do
            if script_data.debug_draw then
                rendering.draw_circle{
                    color = {r=1,g=0.5,b=0.2},
                    radius = 0.5,
                    width = 5,
                    filled = false,
                    target = {ray.x, ray.y},
                    surface = 1,
                    time_to_live = 120,
                }
            end
            
            for j=1,script_data.ray_walk_distance do
                if walk_ray(ray) then
                    table.insert(done, i)
                    break
                end
            end
        end
        
        for i=#done,1,-1 do
            local index = done[i]
            table.remove(script_data.rays, index)
        end
    end
    
    if #script_data.shore_paths > 0 then
        if script_data.current_shore_path > #script_data.shore_paths then
            script_data.current_shore_path = 1
        end
        
        local path = script_data.shore_paths[script_data.current_shore_path]
        if script_data.debug_draw then
            rendering.draw_rectangle{
                color = {r=1,g=0.3,b=0},
                width = 5,
                filled = false;
                left_top = path.position,
                right_bottom = {
                    x = path.position.x + 1,
                    y = path.position.y + 1,
                },
                surface = 1,
                time_to_live = 120,
            }
        end
        
        -- We are at a shore position.  Direction tells us where
        -- we went last to find shore.
        
        local spawn_group = pick_spawn_group()
        local spawnpoints = {}
        local lost = false
        for _=1,spawn_group.spawn_attempts do
            local start_position = path.position
            for _=1,spawn_group.separation do
                if not follow_shore(path) then
                    lost = true
                    break
                end
            end
            
            if lost then
                break
            end
            
            local end_position = path.position
            local spawn = calculate_spawn(
                start_position, end_position, spawn_group.separation
            )
            table.insert(spawnpoints, spawn)
            
            if script_data.debug_draw then
                rendering.draw_circle{
                    color = {r=0,g=0,b=0},
                    radius = 0.5,
                    width = 5,
                    filled = false,
                    target = spawn,
                    surface = 1,
                    time_to_live = 120,
                }
                rendering.draw_line{
                    color = {r=1,g=0.5,b=0.2},
                    width = 5,
                    from = start_position,
                    to = end_position,
                    surface = 1,
                    time_to_live = 120,
                }
            end
        end
        
        if not lost then
            local valid_spawnpoints = {}
            for _,spawn in ipairs(spawnpoints) do
                if game.surfaces[1].can_place_entity{
                    name = "behemoth-biter",
                    position = spawn,
                } then
                    table.insert(valid_spawnpoints, spawn)
                end
            end
            
            if #valid_spawnpoints >= spawn_group.minimum_group_size then
                local middle = math.floor((#valid_spawnpoints + 1) / 2)
                local group = game.surfaces[1].create_unit_group{
                    position = valid_spawnpoints[middle],
                }
            
                for _,spawn in ipairs(valid_spawnpoints) do         
                    local biter = game.surfaces[1].create_entity{
                        name = spawn_group.name,
                        position = spawn,
                    }
                    
                    if biter.valid then
                        local ai = biter.ai_settings
                        ai.allow_destroy_when_commands_fail = true
                        ai.allow_try_return_to_spawner = false
                        group.add_member(biter)
                    end
                end
            
                if #group.members > 0 then
                    group.set_command{
                        type = defines.command.attack_area,
                        destination = script_data.biter_target,
                        radius = 50,
                    }
                    group.start_moving()
                    script_data.unit_groups[group.group_number] = group

                else
                    game.print("No space for group") -- XXX DEBUG
                    group.destroy()
                end
            end
            
            script_data.current_shore_path = script_data.current_shore_path + 1

        -- The shore was lost, maybe a player placed landfill on it.
        -- Convert it back to a ray.
        else
            local index = script_data.current_shore_path
            local position = script_data.shore_paths[index].position
            table.remove(script_data.shore_paths, index)
            
            local length = math.sqrt(
                position.x * position.x + position.y * position.y
            )
            
            -- Nasty corner case
            if length < 0.1 then
                position.x = 1
                position.y = 0
                length = 1
            end
            
            local ray = {
                x = position.x,
                y = position.y,
                dx = position.x / length,
                dy = position.y / length,
            }
            
            table.insert(script_data.rays, ray)
        end
    end
end

local function on_ai_command_completed(event)
    if event.result == defines.behavior_result.fail then
        local group = script_data.unit_groups[event.unit_number]
        if group then
            -- You useless peice of alien goo
            for _,unit in pairs(group.members) do
                unit.destroy()
            end
            group.destroy()
        end
    end
end

-- Heal method
-- XXX Should benchmark this heal vs the ressurect in UPS hit.
local function on_entity_damaged(event)
    local entity = event.entity
    if entity.name == "behemoth-biter" and entity.speed > 0.3 and entity.health < 1000 then
        entity.speed = entity.speed - 0.0001
        entity.health = entity.health + 2000
    end
end


-- Ressurect method
-- Less events, but more work in them
local function on_entity_died(event)
    local entity = event.entity
    if entity.name == "behemoth-biter" and entity.speed > 0.3 then
        local key = entity.position.x .. ", " .. entity.position.y
        script_data.dead_biters[key] = {
            name = "behemoth-biter",
            position = entity.position,
            speed = entity.speed - 0.001,
            orientation = entity.orientation,
        }
        entity.speed = entity.speed - 0.001
        entity.health = 3000
    end
end

local function on_post_entity_died(event)
    local key = event.position.x .. ", " .. event.position.y
    local data = script_data.dead_biters[key]
    if data then
        script_data.dead_biters[key] = nil
        for _, corpse in ipairs(event.corpses) do
            corpse.destroy()
        end
        
        local biter = game.surfaces[event.surface_index].create_entity{
            name = data.name,
            position = data.position,
        }
        
        biter.speed = data.speed
        biter.orientation = data.orientation
    end
end


local poseidon = {}

poseidon.on_configuration_changed = function(data)
    -- TODO: Add migration scripts
end

poseidon.on_init = function()
    log("Poseidon Init")
    global.poseidon_data = global.poseidon_data or script_data
    
    -- XXX DEBUG
    game.forces["player"].research_all_technologies()
    game.forces["player"].technologies["mining-productivity-4"].level = 158
    game.forces["player"].technologies["artillery-shell-range-1"].level = 7
    game.forces["player"].technologies["artillery-shell-speed-1"].level = 6
    game.forces["player"].technologies["refined-flammables-7"].level = 15
    game.forces["player"].technologies["energy-weapons-damage-7"].level = 16
    game.forces["player"].technologies["stronger-explosives-7"].level = 16
    game.forces["player"].technologies["physical-projectile-damage-7"].level = 17
    game.forces["player"].technologies["follower-robot-count-7"].level = 24
    game.forces["player"].technologies["worker-robots-speed-6"].level = 20
    global.chart_distance = 300
    
    
    -- Increase the pathfinder limit so biters keep up
    --game.map_settings.path_finder.max_steps_worked_per_tick = 300
    --game.map_settings.path_finder.max_clients_to_accept_any_new_request = 20
    
    -- Avoid moving entities at a much longer distance, this lets biters
    -- effectively move arround pileups
    --game.map_settings.path_finder.ignore_moving_enemy_collision_distance = 100
    
    -- Hammer it.  The cache is causing far too many issues to be worth the 1ms
    -- per tick that it saves.  In short:
    -- * Units tend to all want to move through the same place causing pileups
    -- * Pileups don't get routed around making them even worse
    game.map_settings.path_finder.use_path_cache = false
    
    -- Destroy everything in the way, not doing so increases the chances of pilup
    --game.map_settings.path_finder.general_entity_collision_penalty = 0
    --game.map_settings.path_finder.general_entity_subsequent_collision_penalty = 0
end

poseidon.on_load = function()
    log("Poseidon Load")

    -- XXX This is what silo script does, but wouldn't this just desync if
    --     if script_data is not in global?
    script_data = global.poseidon_data or script_data
end

poseidon.on_nth_ticks = {
    [1] = on_every_tick,
    [60] = on_gui_tick,
    [24] = on_wrath_tick,
}

local script_events = {
    [defines.events.on_gui_click] = on_gui_click,
    [defines.events.on_player_created] = on_player_created,
    [defines.events.on_ai_command_completed] = on_ai_command_completed,
    [defines.events.on_rocket_launched] = on_rocket_launched,

    -- Biter buffs.  Disabled for now
    -- [defines.events.on_entity_died] = on_entity_died,
    -- [defines.events.on_post_entity_died] = on_post_entity_died,
    -- [defines.events.on_entity_damaged] = on_entity_damaged,
}

poseidon.get_events = function()
    return script_events
end

return poseidon
